@@ -1,11 +1,38 @@
 
 from datetime import datetime, date, time, timedelta
 from fractions import Fraction
+from functools import wraps
 from logging import warning
 from json import JSONEncoder
-from sys import version
+from sys import version, stderr
 from decimal import Decimal
-from .utils import hashodict, call_with_optional_kwargs, NoPandasException, NoNumpyException
+from .utils import hashodict, call_with_optional_kwargs, \
+	get_module_name_from_object, NoEnumException, NoPandasException, \
+	NoNumpyException, str_type
+
+
+def _fallback_wrapper(encoder):
+	"""
+	This decorator makes an encoder run only if the current object hasn't been changed yet.
+	(Changed-ness is checked with is_changed which is based on identity with `id`).
+	"""
+	@wraps(encoder)
+	def fallback_encoder(obj, is_changed, **kwargs):
+		if is_changed:
+			return obj
+		return encoder(obj, is_changed=is_changed, **kwargs)
+	return fallback_encoder
+
+
+def fallback_ignore_unknown(obj, is_changed=None, fallback_value=None):
+	"""
+	This encoder returns None if the object isn't changed by another encoder and isn't a primitive.
+	"""
+	if is_changed:
+		return obj
+	if obj is None or isinstance(obj, (int, float, str_type, bool, list, dict)):
+		return obj
+	return fallback_value
 
 
 class TricksEncoder(JSONEncoder):
@@ -16,14 +43,18 @@ class TricksEncoder(JSONEncoder):
 	Each encoder should make any appropriate changes and return an object,
 	changed or not. This will be passes to the other encoders.
 	"""
-	def __init__(self, obj_encoders=None, silence_typeerror=False, primitives=False, **json_kwargs):
+	def __init__(self, obj_encoders=None, silence_typeerror=False, primitives=False, fallback_encoders=(), **json_kwargs):
 		"""
 		:param obj_encoders: An iterable of functions or encoder instances to try.
-		:param silence_typeerror: If set to True, ignore the TypeErrors that Encoder instances throw (default False).
+		:param silence_typeerror: DEPRECATED - If set to True, ignore the TypeErrors that Encoder instances throw (default False).
 		"""
+		if silence_typeerror and not getattr(TricksEncoder, '_deprecated_silence_typeerror'):
+			TricksEncoder._deprecated_silence_typeerror = True
+			stderr.write('TricksEncoder.silence_typeerror is deprecated and may be removed in a future version\n')
 		self.obj_encoders = []
 		if obj_encoders:
 			self.obj_encoders = list(obj_encoders)
+		self.obj_encoders.extend(_fallback_wrapper(encoder) for encoder in list(fallback_encoders))
 		self.silence_typeerror = silence_typeerror
 		self.primitives = primitives
 		super(TricksEncoder, self).__init__(**json_kwargs)
@@ -44,18 +75,19 @@ class TricksEncoder(JSONEncoder):
 			if hasattr(encoder, 'default'):
 				#todo: write test for this scenario (maybe ClassInstanceEncoder?)
 				try:
-					obj = call_with_optional_kwargs(encoder.default, obj, primitives=self.primitives)
+					obj = call_with_optional_kwargs(encoder.default, obj, primitives=self.primitives, is_changed=id(obj) != prev_id)
 				except TypeError as err:
 					if not self.silence_typeerror:
 						raise
 			elif hasattr(encoder, '__call__'):
-				obj = call_with_optional_kwargs(encoder, obj, primitives=self.primitives)
+				obj = call_with_optional_kwargs(encoder, obj, primitives=self.primitives, is_changed=id(obj) != prev_id)
 			else:
 				raise TypeError('`obj_encoder` {0:} does not have `default` method and is not callable'.format(encoder))
 		if id(obj) == prev_id:
-			#todo: test
-			raise TypeError('Object of type {0:} could not be encoded by {1:} using encoders [{2:s}]'.format(
-				type(obj), self.__class__.__name__, ', '.join(str(encoder) for encoder in self.obj_encoders)))
+			raise TypeError(('Object of type {0:} could not be encoded by {1:} using encoders [{2:s}]. '
+				'You can add an encoders for this type using `extra_obj_encoders`. If you want to \'skip\' this '
+				'object, consider using `fallback_encoders` like `str` or `lambda o: None`.').format(
+					type(obj), self.__class__.__name__, ', '.join(str(encoder) for encoder in self.obj_encoders)))
 		return obj
 
 
@@ -95,6 +127,38 @@ def json_date_time_encode(obj, primitives=False):
 	return dct
 
 
+def enum_instance_encode(obj, primitives=False, with_enum_value=False):
+	"""Encodes an enum instance to json. Note that it can only be recovered if the environment allows the enum to be
+	imported in the same way.
+	:param primitives: If true, encode the enum values as primitive (more readable, but cannot be restored automatically).
+	:param with_enum_value: If true, the value of the enum is also exported (it is not used during import, as it should be constant).
+	"""
+	from enum import Enum
+	if not isinstance(obj, Enum):
+		return obj
+	if primitives:
+		return {obj.name: obj.value}
+	mod = get_module_name_from_object(obj)
+	representation = dict(
+		__enum__=dict(
+			# Don't use __instance_type__ here since enums members cannot be created with __new__
+			# Ie we can't rely on class deserialization to read them.
+			__enum_instance_type__=[mod, type(obj).__name__],
+			name=obj.name,
+		),
+	)
+	if with_enum_value:
+		representation['__enum__']['value'] = obj.value
+	return representation
+
+
+def noenum_instance_encode(obj, primitives=False):
+	if type(obj.__class__).__name__ == 'EnumMeta':
+		raise NoEnumException(('Trying to encode an object of type {0:} which appears to be '
+			'an enum, but enum support is not enabled, perhaps it is not installed.').format(type(obj)))
+	return obj
+
+
 def class_instance_encode(obj, primitives=False):
 	"""
 	Encodes a class instance to json. Note that it can only be recovered if the environment allows the class to be
@@ -112,12 +176,10 @@ def class_instance_encode(obj, primitives=False):
 		except TypeError:
 			raise TypeError(('instance "{0:}" of class "{1:}" cannot be encoded because it\'s __new__ method '
 				'cannot be called, perhaps it requires extra parameters').format(obj, obj.__class__))
-		mod = obj.__class__.__module__
-		if mod == '__main__':
-			mod = None
-			warning(('class {0:} seems to have been defined in the main file; unfortunately this means'
-				' that it\'s module/import path is unknown, so you might have to provide cls_lookup_map when '
-				'decoding').format(obj.__class__))
+		mod = get_module_name_from_object(obj)
+		if mod == 'threading':
+			# In Python2, threading objects get serialized, which is probably unsafe
+			return obj
 		name = obj.__class__.__name__
 		if hasattr(obj, '__json_encode__'):
 			attrs = obj.__json_encode__()
